# -*- coding: utf-8 -*-
"""pyeasyga

A simple and easy-to-use genetic algorithm implementation library in Python.

For a bit array solution representation, simply instantiate the
GeneticAlgorithm class with input data, define and supply a fitness function,
run the Genetic Algorithm, and retrieve the solution!

Other solution representations will require setting some more attributes.

"""

__author__ = 'Ayodeji Remi-Omosowon'
__email__ = 'remiomosowon@gmail.com'
__version__ = '0.2.0'
